#include <syslog.h>
#include <stdio.h>

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_request.h"
#include "ap_provider.h"

#include <apr_general.h>
#include <apr_network_io.h>
#include <apr_strings.h>

#include "mod_auth.h"

/*
 * Copyright 2008 ed neville 
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * http://www.s5h.net/
 *
 * $Id: mod_authn_pop_basic.c 163 2008-08-03 22:15:09Z ed $
 * $URL: svn+ssh://fw/repos/code/apache2_modules/pop/mod_authn_pop_basic.c $
 */

/*
 * possible return values
 * AUTH_GENERAL_ERROR
 * AUTH_DENIED
 * AUTH_GRANTED
 */

#define BUFFER 8192

module AP_MODULE_DECLARE_DATA authn_pop_module;

typedef struct pop_cfg {
	char *config_pop_server_address; /* pop server address */
	int config_pop_port_number;
} pop_cfg;

static pop_cfg *our_dir_config( const request_rec *r )
{
	return( ( pop_cfg* )ap_get_module_config( r->per_dir_config, &authn_pop_module ) );
}

static void *pop_dir_conf( apr_pool_t *pool, char *x ) {
	void *dir = (void*)apr_pcalloc( pool, sizeof( pop_cfg ) );
	return( (void*) dir );
}

static int authn_pop_read_login_banner( request_rec *r, apr_socket_t *s ) {
	int got_ok = APR_ENOSPC;
	apr_size_t bytes_read = 0;
	apr_status_t status;
	char *buf = apr_palloc( r->pool, BUFFER );
	int retry_count;
	
	bytes_read = BUFFER;

	status = apr_socket_recv( s, buf, &bytes_read );

	if( status != APR_SUCCESS ) {
		return( APR_ENOTSOCK );
	}
	
	if( strncmp( buf, "+OK", 3 ) == 0 ) {
		return( APR_SUCCESS );
	}

	return( APR_ENOTSOCK );
}

static int authn_pop_send_login( request_rec *r, apr_socket_t *s, const char *user, const char *password ) {
	
	// allocate way more than we really need
	apr_status_t status;
	char *login_string;
	apr_size_t len;
	apr_size_t bytes_read;
	int i;
	
	login_string = apr_pstrcat( r->pool, "USER ", user, "\r\nPASS ", password, "\r\n", NULL );

	if( login_string == NULL ) {
		return( AUTH_GENERAL_ERROR );
	}

	// now we try and login to the pop server with the details that we were given a moment ago

	len = strlen( login_string );

	status = apr_socket_send( s, login_string, &len );

	if( status != APR_SUCCESS ) {
		return( AUTH_GENERAL_ERROR );
	}


	for( i == 0 ; i<2 ; i++ ) {
		bytes_read = BUFFER;

		status = apr_socket_recv( s, login_string, &bytes_read );

		if( status != APR_SUCCESS ) {
			return( APR_ENOTSOCK );
		}

		if( strncmp( login_string, "+OK", 3 ) != 0 ) {
			return( AUTH_DENIED );
		}
	}

	return( AUTH_GRANTED );
}

static authn_status authn_pop( request_rec *r, const char *user, const char *password ) {

	apr_sockaddr_t *sa;
	apr_status_t status;
	apr_socket_t *s;
	apr_size_t len;
	int family = APR_INET;
	pop_cfg *cfg;

	cfg = our_dir_config( r );
	if( cfg == NULL ) {
		return( AUTH_DENIED );
	}

	if( cfg->config_pop_server_address == NULL ) {
		return( AUTH_DENIED );
	}

	status = apr_sockaddr_info_get( &sa, cfg->config_pop_server_address, APR_INET, cfg->config_pop_port_number, 0, r->pool );
	if( status != APR_SUCCESS ) {
		/*ap_log_error(APLOG_MARK, APLOG_ERR, status, r->server,
                         "Could not create socket addr " );*/
		return( AUTH_GENERAL_ERROR );
	}

	status = apr_socket_create( &s, sa->family, SOCK_STREAM, APR_PROTO_TCP, r->pool );

	if( status != APR_SUCCESS ) {
		/* ap_log_error(APLOG_MARK, APLOG_ERR, status, r->server,
                         "Could not create socket" );*/
		return( AUTH_GENERAL_ERROR );
	}

	status = apr_socket_connect( s, sa );
	if( status != APR_SUCCESS ) {
		return( AUTH_GENERAL_ERROR );
	}

	/* set up some defaults. we don't want the socket to block, but we will be happy to wait for one second for data */

	apr_socket_opt_set( s, APR_SO_NONBLOCK, 1 );
	apr_socket_timeout_set( s, 100000 );

	/* read the welcome banner - return auth failure if we don't get a 2xx banner */
	if( authn_pop_read_login_banner( r, s ) != APR_SUCCESS ) {
		return( AUTH_GENERAL_ERROR );
	}

	if( authn_pop_send_login( r, s, user, password ) == APR_SUCCESS ) {
		return( AUTH_GRANTED );
	}
	
	return( AUTH_DENIED );
}

static const char *set_pop_server_address( cmd_parms *cmd, void *cfg, const char *val ) {
	if( cfg == NULL ) {
		return( NULL );
	}
	((pop_cfg*)cfg)->config_pop_server_address = val;
	return( NULL );
}

static const char *set_pop_server_port( cmd_parms *cmd, void *cfg, const char *val ) {
	int port_number;
	apr_int64_t i64;

	if( cfg == NULL ) {
		return( NULL );
	}
	
	i64 = apr_strtoi64( val, NULL, 10 );
	port_number = i64 & 0xFF;
	((pop_cfg*)cfg)->config_pop_port_number = port_number;
	
	return( NULL );
}

static const authn_provider authn_pop_provider = { &authn_pop, NULL };

static void register_hooks( apr_pool_t *p ) {
	ap_register_provider( p, AUTHN_PROVIDER_GROUP, "vchkpw", "0", &authn_pop_provider );
}


static const command_rec pop_cmds[] =  {
	AP_INIT_TAKE1(
		"POP_Server",
		set_pop_server_address,
		NULL,
		OR_ALL,
		"pop server to authenticate with" ),
	AP_INIT_TAKE1(
		"POP_Port",
		set_pop_server_port,
		NULL,
		OR_ALL,
		"pop port number used in authentication" ),
	{NULL}

};


module AP_MODULE_DECLARE_DATA authn_pop_module = {
	STANDARD20_MODULE_STUFF,
	pop_dir_conf,	/* dir config creater */
	NULL,		/* dir merger --- default is to override */
	NULL,		/* server config */
	NULL,		/* merge server config */
	pop_cmds,	/* command apr_table_t */
	register_hooks	/* register hooks */
};
